__version__ = "0.2.5"
__git_commiter_name__ = "Tal Yalon"
__git_commiter_email__ = "yalon@infinidat.com"
__git_branch__ = '(detached from v0.2.5)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = 'b8dcdc8cef115f50d72f514446f877aca51ab8d6'
__git_head_subject__ = 'fixed timestamp issue and rsyslog not handling comma as the second fraction separator well'
__git_head_message__ = ''
__git_dirty_diff__ = ''